#include "Cv8_main.h"

//---------------------
//----    SCENE    ----
//---------------------

/// Reloads all shaders
void reload_shaders()
{
	// Initialize shader programs
	//
	// The rendering is getting more and more sophisticated, so we need more and more shaders.
	// We use the following shaders:
	//		notexture_*_program		- objects without textures, only with a single color
	//		texture_*_program		- objects with textures
	//		skybox_*_program		- skybox in the background
	//
	//		*_{nothing}_program					- shaders for simple rendering as we did in the last lectures
	//		*_to_cube_program					- shaders that render all six faces of the cube map simultaneously
	//		*_to_cube_invocations_program		- shaders that render all six faces of the cube map simultaneously and use invocations in geometry shaders
	//		*_to_cube_instancing_program		- shaders that render all six faces of the cube map simultaneously and use instancing with geometry shaders
	//		*_to_cube_instancing_no_gs_program	- shaders that render all six faces of the cube map simultaneously and use instancing without geometry shaders
	//
	//		display_cubemap_program		- for the central object, displays the cube map without reflections
	//		reflection_program			- for the central object, renders the object with texture and with reflections of the environment
	//
	// In addition to that, we use ShaderProgramMap to manage the shaders of notexture_*, texture_* and skybox_*:
	//		RENDER_STANDARD					- use *_{nothing}_program
	//		RENDER_TO_CUBE					- use *_to_cube_program
	//		RENDER_TO_CUBE_INVOCATIONS		- use *_to_cube_invocations_program
	//		RENDER_TO_CUBE_INSTANCING		- use *_to_cube_instancing_program
	//		RENDER_TO_CUBE_INSTANCING_NO_GS	- use *_to_cube_instancing_no_gs_program
	//
	// And final note, to decrease the number of shaders, we use the vertex shader for objects with textures also for
	// objects without textures. The fragment shader of objects without textures just ignores the texture coordinates.

	// Shaders for simple rendering as we did in the last lectures
	notexture_program.Init();
	notexture_program.AddVertexShader("Shaders/texture_vertex.glsl");
	notexture_program.AddFragmentShader("Shaders/notexture_fragment.glsl");
	notexture_program.Link();

	texture_program.Init();
	texture_program.AddVertexShader("Shaders/texture_vertex.glsl");
	texture_program.AddFragmentShader("Shaders/texture_fragment.glsl");
	texture_program.Link();

	skybox_program.Init();
	skybox_program.AddVertexShader("Shaders/skybox_vertex.glsl");
	skybox_program.AddFragmentShader("Shaders/skybox_fragment.glsl");
	skybox_program.Link();

	// Shaders that render all six faces of the cube map simultaneously
	texture_to_cube_program.Init();
	texture_to_cube_program.AddVertexShader("Shaders/texture_to_cube_vertex.glsl");
	texture_to_cube_program.AddGeometryShader("Shaders/texture_to_cube_geometry.glsl");
	texture_to_cube_program.AddFragmentShader("Shaders/texture_fragment.glsl");
	texture_to_cube_program.Link();

	notexture_to_cube_program.Init();
	notexture_to_cube_program.AddVertexShader("Shaders/texture_to_cube_vertex.glsl");
	notexture_to_cube_program.AddGeometryShader("Shaders/texture_to_cube_geometry.glsl");
	notexture_to_cube_program.AddFragmentShader("Shaders/notexture_fragment.glsl");
	notexture_to_cube_program.Link();

	skybox_to_cube_program.Init();
	skybox_to_cube_program.AddVertexShader("Shaders/skybox_to_cube_vertex.glsl");
	skybox_to_cube_program.AddGeometryShader("Shaders/skybox_to_cube_geometry.glsl");
	skybox_to_cube_program.AddFragmentShader("Shaders/skybox_fragment.glsl");
	skybox_to_cube_program.Link();

	// Shaders that render all six faces of the cube map simultaneously and use invocations in geometry shaders
	skybox_to_cube_invocations_program.Init();
	skybox_to_cube_invocations_program.AddVertexShader("Shaders/skybox_to_cube_vertex.glsl");
	skybox_to_cube_invocations_program.AddGeometryShader("Shaders/skybox_to_cube_invocations_geometry.glsl");
	skybox_to_cube_invocations_program.AddFragmentShader("Shaders/skybox_fragment.glsl");
	skybox_to_cube_invocations_program.Link();

	texture_to_cube_invocations_program.Init();
	texture_to_cube_invocations_program.AddVertexShader("Shaders/texture_to_cube_vertex.glsl");
	texture_to_cube_invocations_program.AddGeometryShader("Shaders/texture_to_cube_invocations_geometry.glsl");
	texture_to_cube_invocations_program.AddFragmentShader("Shaders/texture_fragment.glsl");
	texture_to_cube_invocations_program.Link();

	notexture_to_cube_invocations_program.Init();
	notexture_to_cube_invocations_program.AddVertexShader("Shaders/texture_to_cube_vertex.glsl");
	notexture_to_cube_invocations_program.AddGeometryShader("Shaders/texture_to_cube_invocations_geometry.glsl");
	notexture_to_cube_invocations_program.AddFragmentShader("Shaders/notexture_fragment.glsl");
	notexture_to_cube_invocations_program.Link();

	// Shaders that render all six faces of the cube map simultaneously and use instancing with geometry shaders
	skybox_to_cube_instancing_program.Init();
	skybox_to_cube_instancing_program.AddVertexShader("Shaders/skybox_to_cube_instancing_vertex.glsl");
	skybox_to_cube_instancing_program.AddGeometryShader("Shaders/skybox_to_cube_instancing_geometry.glsl");
	skybox_to_cube_instancing_program.AddFragmentShader("Shaders/skybox_fragment.glsl");
	skybox_to_cube_instancing_program.Link();

	texture_to_cube_instancing_program.Init();
	texture_to_cube_instancing_program.AddVertexShader("Shaders/texture_to_cube_instancing_vertex.glsl");
	texture_to_cube_instancing_program.AddGeometryShader("Shaders/texture_to_cube_instancing_geometry.glsl");
	texture_to_cube_instancing_program.AddFragmentShader("Shaders/texture_fragment.glsl");
	texture_to_cube_instancing_program.Link();

	notexture_to_cube_instancing_program.Init();
	notexture_to_cube_instancing_program.AddVertexShader("Shaders/texture_to_cube_instancing_vertex.glsl");
	notexture_to_cube_instancing_program.AddGeometryShader("Shaders/texture_to_cube_instancing_geometry.glsl");
	notexture_to_cube_instancing_program.AddFragmentShader("Shaders/notexture_fragment.glsl");
	notexture_to_cube_instancing_program.Link();

	// Shaders that render all six faces of the cube map simultaneously and use instancing without geometry shaders
	if (IsOpenGLExtensionPresent("GL_ARB_shader_viewport_layer_array"))		// We need this extension
	{
		skybox_to_cube_instancing_no_gs_program.Init();
		skybox_to_cube_instancing_no_gs_program.AddVertexShader("Shaders/skybox_to_cube_instancing_no_gs_vertex.glsl");
		skybox_to_cube_instancing_no_gs_program.AddFragmentShader("Shaders/skybox_fragment.glsl");
		skybox_to_cube_instancing_no_gs_program.Link();

		texture_to_cube_instancing_no_gs_program.Init();
		texture_to_cube_instancing_no_gs_program.AddVertexShader("Shaders/texture_to_cube_instancing_no_gs_vertex.glsl");
		texture_to_cube_instancing_no_gs_program.AddFragmentShader("Shaders/texture_fragment.glsl");
		texture_to_cube_instancing_no_gs_program.Link();

		notexture_to_cube_instancing_no_gs_program.Init();
		notexture_to_cube_instancing_no_gs_program.AddVertexShader("Shaders/texture_to_cube_instancing_no_gs_vertex.glsl");
		notexture_to_cube_instancing_no_gs_program.AddFragmentShader("Shaders/notexture_fragment.glsl");
		notexture_to_cube_instancing_no_gs_program.Link();
	}
	else
	{
		// The extension is not present, we cannot load these shaders
		skybox_to_cube_instancing_no_gs_program.Destroy();
		texture_to_cube_instancing_no_gs_program.Destroy();
		notexture_to_cube_instancing_no_gs_program.Destroy();
	}

	// For the central object, displays the cube map without reflections
	display_cubemap_program.Init();
	display_cubemap_program.AddVertexShader("Shaders/display_cubemap_vertex.glsl");
	display_cubemap_program.AddFragmentShader("Shaders/skybox_fragment.glsl");
	display_cubemap_program.Link();

	// For the central object, renders the object with texture and with reflections of the environment
	reflection_program.Init();
	reflection_program.AddVertexShader("Shaders/texture_vertex.glsl");
	reflection_program.AddFragmentShader("Shaders/reflection_fragment.glsl");
	reflection_program.Link();

	// Setup the shader maps
	notexture_program_map.RemoveAll();
	notexture_program_map.AddProgram(RENDER_STANDARD, &notexture_program);
	notexture_program_map.AddProgram(RENDER_TO_CUBE, &notexture_to_cube_program);
	notexture_program_map.AddProgram(RENDER_TO_CUBE_INVOCATIONS, &notexture_to_cube_invocations_program);
	notexture_program_map.AddProgram(RENDER_TO_CUBE_INSTANCING, &notexture_to_cube_instancing_program);
	notexture_program_map.AddProgram(RENDER_TO_CUBE_INSTANCING_NO_GS, &notexture_to_cube_instancing_no_gs_program);
	
	texture_program_map.RemoveAll();
	texture_program_map.AddProgram(RENDER_STANDARD, &texture_program);
	texture_program_map.AddProgram(RENDER_TO_CUBE, &texture_to_cube_program);
	texture_program_map.AddProgram(RENDER_TO_CUBE_INVOCATIONS, &texture_to_cube_invocations_program);
	texture_program_map.AddProgram(RENDER_TO_CUBE_INSTANCING, &texture_to_cube_instancing_program);
	texture_program_map.AddProgram(RENDER_TO_CUBE_INSTANCING_NO_GS, &texture_to_cube_instancing_no_gs_program);
	
	skybox_program_map.RemoveAll();
	skybox_program_map.AddProgram(RENDER_STANDARD, &skybox_program);
	skybox_program_map.AddProgram(RENDER_TO_CUBE, &skybox_to_cube_program);
	skybox_program_map.AddProgram(RENDER_TO_CUBE_INVOCATIONS, &skybox_to_cube_invocations_program);
	skybox_program_map.AddProgram(RENDER_TO_CUBE_INSTANCING, &skybox_to_cube_instancing_program);
	skybox_program_map.AddProgram(RENDER_TO_CUBE_INSTANCING_NO_GS, &skybox_to_cube_instancing_no_gs_program);

	cout << "Shaders are reloaded" << endl;
}

/// Initializes all objects of our scene
void init_scene()
{
	//----------------------------------------------
	//--  Display some information about the OpenGL

	PrintOpenGLInfo();

	if (IsOpenGLExtensionPresent("GL_ARB_shader_viewport_layer_array"))
		cout << "Extension GL_ARB_shader_viewport_layer_array is present" << endl;
	else
		cout << "Extension GL_ARB_shader_viewport_layer_array is not present" << endl;

	cout << endl;

	//----------------------------------------------
	//--  Load shaders

	reload_shaders();
	
	//----------------------------------------------
	//--  Prepare the lights

	// Create only a single directional light which direction corresponds to the position of the sun in the cubemap
	PhongLights_ubo.Init();
	PhongLights_ubo.SetGlobalAmbient(glm::vec3(0.2f));
	PhongLights_ubo.PhongLights.clear();
	PhongLights_ubo.PhongLights.push_back(
		PhongLight::CreateDirectionalLight(
		glm::vec3(1.0f, 1.0f, 1.0f),
		glm::vec3(0.0f), glm::vec3(1.0f), glm::vec3(1.0f)));
	PhongLights_ubo.UpdateOpenGLData();

	//----------------------------------------------
	//--  Prepare materials

	// Prepare materials for the random scene and for the floor
	RedMaterial_ubo.Init();
	GreenMaterial_ubo.Init();
	BlueMaterial_ubo.Init();
	CyanMaterial_ubo.Init();
	MagentaMaterial_ubo.Init();
	YellowMaterial_ubo.Init();
	WhiteMaterial_ubo.Init();
	FloorMaterial_ubo.Init();
	RedMaterial_ubo.SetMaterial		(PhongMaterial::CreateBasicMaterial(glm::vec3(1.0f, 0.0f, 0.0f), true, 200.0f, 1.0f));
	GreenMaterial_ubo.SetMaterial	(PhongMaterial::CreateBasicMaterial(glm::vec3(0.0f, 1.0f, 0.0f), true, 200.0f, 1.0f));
	BlueMaterial_ubo.SetMaterial	(PhongMaterial::CreateBasicMaterial(glm::vec3(0.0f, 0.0f, 1.0f), true, 200.0f, 1.0f));
	CyanMaterial_ubo.SetMaterial	(PhongMaterial::CreateBasicMaterial(glm::vec3(0.0f, 1.0f, 1.0f), true, 200.0f, 1.0f));
	MagentaMaterial_ubo.SetMaterial	(PhongMaterial::CreateBasicMaterial(glm::vec3(1.0f, 0.0f, 1.0f), true, 200.0f, 1.0f));
	YellowMaterial_ubo.SetMaterial	(PhongMaterial::CreateBasicMaterial(glm::vec3(1.0f, 1.0f, 0.0f), true, 200.0f, 1.0f));
	WhiteMaterial_ubo.SetMaterial	(PhongMaterial::CreateBasicMaterial(glm::vec3(1.0f, 1.0f, 1.0f), true, 200.0f, 1.0f));
	FloorMaterial_ubo.SetMaterial	(PhongMaterial::CreateBasicMaterial(glm::vec3(0.7f, 0.7f, 0.7f), true, 200.0f, 1.0f));
	RedMaterial_ubo.UpdateOpenGLData();
	GreenMaterial_ubo.UpdateOpenGLData();
	BlueMaterial_ubo.UpdateOpenGLData();
	CyanMaterial_ubo.UpdateOpenGLData();
	MagentaMaterial_ubo.UpdateOpenGLData();
	YellowMaterial_ubo.UpdateOpenGLData();
	WhiteMaterial_ubo.UpdateOpenGLData();
	FloorMaterial_ubo.UpdateOpenGLData();

	// Add the materials into our list of materials (we do not add the material for the floor, only the basic colors)
	Colors_ubo.push_back(&RedMaterial_ubo);
	Colors_ubo.push_back(&GreenMaterial_ubo);
	Colors_ubo.push_back(&BlueMaterial_ubo);
	Colors_ubo.push_back(&CyanMaterial_ubo);
	Colors_ubo.push_back(&MagentaMaterial_ubo);
	Colors_ubo.push_back(&YellowMaterial_ubo);
	Colors_ubo.push_back(&WhiteMaterial_ubo);

	//----------------------------------------------
	//--  Prepare all cameras
	
	// Create the basic camera through which we look into the scene
	the_camera.SetCenter(glm::vec3(0.0f));
	the_camera.SetCamera(-0.5f, -0.5f, 15.0f);
	CameraData_ubo.Init();

	// Create six separate cameras to use when rendering six faces of the cube map
	CubemapSeparateCameraData_ubo[0].Init();
	CubemapSeparateCameraData_ubo[1].Init();
	CubemapSeparateCameraData_ubo[2].Init();
	CubemapSeparateCameraData_ubo[3].Init();
	CubemapSeparateCameraData_ubo[4].Init();
	CubemapSeparateCameraData_ubo[5].Init();
	// All six cameras have the same projection
	CubemapSeparateCameraData_ubo[0].SetProjection(glm::perspective(glm::radians(90.0f), 1.0f, 0.1f, 1000.0f));
	CubemapSeparateCameraData_ubo[1].SetProjection(glm::perspective(glm::radians(90.0f), 1.0f, 0.1f, 1000.0f));
	CubemapSeparateCameraData_ubo[2].SetProjection(glm::perspective(glm::radians(90.0f), 1.0f, 0.1f, 1000.0f));
	CubemapSeparateCameraData_ubo[3].SetProjection(glm::perspective(glm::radians(90.0f), 1.0f, 0.1f, 1000.0f));
	CubemapSeparateCameraData_ubo[4].SetProjection(glm::perspective(glm::radians(90.0f), 1.0f, 0.1f, 1000.0f));
	CubemapSeparateCameraData_ubo[5].SetProjection(glm::perspective(glm::radians(90.0f), 1.0f, 0.1f, 1000.0f));
	// Each camera looks in specific direction (POSITIVE/NEGATIVE_X/Y/Z) and is properly oriented
	CubemapSeparateCameraData_ubo[0].SetCamera(glm::lookAt(glm::vec3(0.0f), glm::vec3(+1.0f, 0.0f, 0.0f), glm::vec3(0.0f, -1.0f, 0.0f)));
	CubemapSeparateCameraData_ubo[1].SetCamera(glm::lookAt(glm::vec3(0.0f), glm::vec3(-1.0f, 0.0f, 0.0f), glm::vec3(0.0f, -1.0f, 0.0f)));
	CubemapSeparateCameraData_ubo[2].SetCamera(glm::lookAt(glm::vec3(0.0f), glm::vec3(0.0f, +1.0f, 0.0f), glm::vec3(0.0f, 0.0f, 1.0f)));
	CubemapSeparateCameraData_ubo[3].SetCamera(glm::lookAt(glm::vec3(0.0f), glm::vec3(0.0f, -1.0f, 0.0f), glm::vec3(0.0f, 0.0f, -1.0f)));
	CubemapSeparateCameraData_ubo[4].SetCamera(glm::lookAt(glm::vec3(0.0f), glm::vec3(0.0f, 0.0f, +1.0f), glm::vec3(0.0f, -1.0f, 0.0f)));
	CubemapSeparateCameraData_ubo[5].SetCamera(glm::lookAt(glm::vec3(0.0f), glm::vec3(0.0f, 0.0f, -1.0f), glm::vec3(0.0f, -1.0f, 0.0f)));
	// Update OpenGL buffer
	CubemapSeparateCameraData_ubo[0].UpdateOpenGLData();
	CubemapSeparateCameraData_ubo[1].UpdateOpenGLData();
	CubemapSeparateCameraData_ubo[2].UpdateOpenGLData();
	CubemapSeparateCameraData_ubo[3].UpdateOpenGLData();
	CubemapSeparateCameraData_ubo[4].UpdateOpenGLData();
	CubemapSeparateCameraData_ubo[5].UpdateOpenGLData();

	// Create a group of six cameras to use when rendering six faces of the cube map
	CubemapCameraData_ubo.Init(6);
	// All six cameras have the same projection
	CubemapCameraData_ubo.SetProjection(0, glm::perspective(glm::radians(90.0f), 1.0f, 0.1f, 1000.0f));
	CubemapCameraData_ubo.SetProjection(1, glm::perspective(glm::radians(90.0f), 1.0f, 0.1f, 1000.0f));
	CubemapCameraData_ubo.SetProjection(2, glm::perspective(glm::radians(90.0f), 1.0f, 0.1f, 1000.0f));
	CubemapCameraData_ubo.SetProjection(3, glm::perspective(glm::radians(90.0f), 1.0f, 0.1f, 1000.0f));
	CubemapCameraData_ubo.SetProjection(4, glm::perspective(glm::radians(90.0f), 1.0f, 0.1f, 1000.0f));
	CubemapCameraData_ubo.SetProjection(5, glm::perspective(glm::radians(90.0f), 1.0f, 0.1f, 1000.0f));
	// Each camera looks in specific direction (POSITIVE/NEGATIVE_X/Y/Z) and is properly oriented
	CubemapCameraData_ubo.SetCamera(0, glm::lookAt(glm::vec3(0.0f), glm::vec3(+1.0f, 0.0f, 0.0f), glm::vec3(0.0f, -1.0f, 0.0f)));
	CubemapCameraData_ubo.SetCamera(1, glm::lookAt(glm::vec3(0.0f), glm::vec3(-1.0f, 0.0f, 0.0f), glm::vec3(0.0f, -1.0f, 0.0f)));
	CubemapCameraData_ubo.SetCamera(2, glm::lookAt(glm::vec3(0.0f), glm::vec3(0.0f, +1.0f, 0.0f), glm::vec3(0.0f, 0.0f, 1.0f)));
	CubemapCameraData_ubo.SetCamera(3, glm::lookAt(glm::vec3(0.0f), glm::vec3(0.0f, -1.0f, 0.0f), glm::vec3(0.0f, 0.0f, -1.0f)));
	CubemapCameraData_ubo.SetCamera(4, glm::lookAt(glm::vec3(0.0f), glm::vec3(0.0f, 0.0f, +1.0f), glm::vec3(0.0f, -1.0f, 0.0f)));
	CubemapCameraData_ubo.SetCamera(5, glm::lookAt(glm::vec3(0.0f), glm::vec3(0.0f, 0.0f, -1.0f), glm::vec3(0.0f, -1.0f, 0.0f)));
	// Update OpenGL buffer
	CubemapCameraData_ubo.UpdateOpenGLData();

	//----------------------------------------------
	//--  Prepare geometries
	
	geom_cube = CreateCube();
	geom_sphere = CreateSphere();
	geom_torus = CreateTorus();
	geom_cylinder = CreateCylinder();
	geom_capsule = CreateCapsule();
	geom_teapot = CreateTeapot();
	geom_fullscreen_quad = CreateFullscreenQuad();

	// Add the geometries into our list of geometries. Also add matrices that describe
	// how to transform the objects so that they "sit" on xz plane.
	Geometries.push_back(make_pair(&geom_cube,
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 1.0f, 0.0f))));
	Geometries.push_back(make_pair(&geom_sphere,
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 1.0f, 0.0f))));
	Geometries.push_back(make_pair(&geom_torus,
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 1.5f, 0.0f)) *
		glm::rotate(glm::mat4(1.0f), glm::radians(90.0f), glm::vec3(1.0f, 0.0f, 0.0f))));
	Geometries.push_back(make_pair(&geom_cylinder,
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 0.5f, 0.0f)) *
		glm::rotate(glm::mat4(1.0f), glm::radians(90.0f), glm::vec3(1.0f, 0.0f, 0.0f))));
	Geometries.push_back(make_pair(&geom_capsule,
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 0.5f, 0.0f)) *
		glm::rotate(glm::mat4(1.0f), glm::radians(90.0f), glm::vec3(1.0f, 0.0f, 0.0f))));
	Geometries.push_back(make_pair(&geom_teapot,
		glm::mat4(1.0f)));

	//----------------------------------------------
	//--  Load textures and set their parameters

	// Load wood texture
	wood_tex = CreateAndLoadTexture2D(L"../../textures/wood.jpg");
	glBindTexture(GL_TEXTURE_2D, wood_tex);
	SetTexture2DParameters(GL_REPEAT, GL_REPEAT, GL_LINEAR, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);

	// Load lenna texture
	lenna_tex = CreateAndLoadTexture2D(L"../../textures/Lenna.jpg");
	glBindTexture(GL_TEXTURE_2D, lenna_tex);
	SetTexture2DParameters(GL_REPEAT, GL_REPEAT, GL_LINEAR, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);

	// Load meadow skybox
	meadow_skybox_tex = CreateAndLoadTextureCube(
		L"../../textures/meadow_skybox/meadow_skybox_px.jpg",
		L"../../textures/meadow_skybox/meadow_skybox_nx.jpg",
		L"../../textures/meadow_skybox/meadow_skybox_py.jpg",
		L"../../textures/meadow_skybox/meadow_skybox_ny.jpg",
		L"../../textures/meadow_skybox/meadow_skybox_pz.jpg",
		L"../../textures/meadow_skybox/meadow_skybox_nz.jpg");
	glBindTexture(GL_TEXTURE_CUBE_MAP, meadow_skybox_tex);
	SetTextureCubeParameters(GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_LINEAR, GL_LINEAR);
	glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
	
	// Load miramar skybox
	miramar_skybox_tex = CreateAndLoadTextureCube(
		L"../../textures/miramar_skybox/miramar_skybox_px.jpg",
		L"../../textures/miramar_skybox/miramar_skybox_nx.jpg",
		L"../../textures/miramar_skybox/miramar_skybox_py.jpg",
		L"../../textures/miramar_skybox/miramar_skybox_ny.jpg",
		L"../../textures/miramar_skybox/miramar_skybox_pz.jpg",
		L"../../textures/miramar_skybox/miramar_skybox_nz.jpg");
	glBindTexture(GL_TEXTURE_CUBE_MAP, miramar_skybox_tex);
	SetTextureCubeParameters(GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_LINEAR, GL_LINEAR);
	glBindTexture(GL_TEXTURE_CUBE_MAP, 0);

	// Choose initial skybox texture
	current_skybox_tex = meadow_skybox_tex;
	
	// Add the textures into our list of textures
	Textures.push_back(wood_tex);
	Textures.push_back(lenna_tex);

	//----------------------------------------------
	//--  Create a random scene

	// We create a grid of random objects with random materials.
	srand(12345);
	glm::vec3 xyz_start = glm::vec3(-20.0f, -5.0f, -20.0f);
	glm::vec3 xyz_spacing = glm::vec3(5.0f);
	glm::ivec3 xyz_grid_size = glm::ivec3(9, 1, 9);
	glm::ivec3 xyz_grid_empty_size = glm::ivec3(3, 1, 3);
	glm::ivec3 xyz_grid_empty_start = (xyz_grid_size - xyz_grid_empty_size) / 2;

	Models_ubo.resize(xyz_grid_size.x * xyz_grid_size.y * xyz_grid_size.z -
		xyz_grid_empty_size.x * xyz_grid_empty_size.y * xyz_grid_empty_size.z);

	int model_idx = 0;
	for (int x = 0; x < xyz_grid_size.x; x++)
	for (int y = 0; y < xyz_grid_size.y; y++)
	for (int z = 0; z < xyz_grid_size.z; z++)
	{
		glm::ivec3 xyz = glm::ivec3(x, y, z);
		if (glm::all(glm::greaterThanEqual(xyz, xyz_grid_empty_start)) &&
			glm::all(glm::lessThan(xyz, xyz_grid_empty_start + xyz_grid_empty_size)))
			continue;

		// Choose the geometry randomly
		int geometry = rand() % int(Geometries.size());

		// Rotate the object randomly around the y axis
		glm::mat4 rotation = glm::rotate(glm::mat4(1.0f), float(rand()) / float(RAND_MAX) * float(M_PI) * 2.0f, glm::vec3(0.0f, 1.0f, 0.0f));

		// Place the object at the proper place in the grid
		glm::mat4 translation = glm::translate(glm::mat4(1.0f), xyz_start + xyz_spacing * glm::vec3(xyz));

		// Set the size of the object
		glm::mat4 scale = glm::scale(glm::mat4(1.0f), glm::vec3(1.2f));

		// Choose the material randomly, choose randomly between the colors and the textures
		int material = rand() % int(Colors_ubo.size() + Textures.size());

		// Compute the model matrix
		Models_ubo[model_idx].Init();
		Models_ubo[model_idx].SetMatrix(translation * rotation * scale * Geometries[geometry].second);
		Models_ubo[model_idx].UpdateOpenGLData();

		// Create the scene object and add it into the list
		SceneObject scene_object;
		scene_object.geometry = Geometries[geometry].first;
		scene_object.model_ubo = &Models_ubo[model_idx];
		if (material < int(Colors_ubo.size()))
		{
			// Object with a color without textures
			scene_object.programs = &notexture_program_map;
			scene_object.material_ubo = Colors_ubo[material];
			scene_object.texture = 0;
		}
		else
		{
			// Object with a texture
			scene_object.programs = &texture_program_map;
			scene_object.material_ubo = &WhiteMaterial_ubo;
			scene_object.texture = Textures[material - int(Colors_ubo.size())];
		}
		ObjectsInScene.push_back(scene_object);
		model_idx++;
	}

	// Add also the floor, the rotating objects in corners and the orbiting object

	// Prepare the objects' model matrices. Set the model matrix of the floor, the model matrices of other objects
	// are set in update (they are animated)
	FloorModel_ubo.Init();
	CornerObjectsModel_ubo[0].Init();
	CornerObjectsModel_ubo[1].Init();
	CornerObjectsModel_ubo[2].Init();
	CornerObjectsModel_ubo[3].Init();
	OrbitingObjectModel_ubo.Init();

	FloorModel_ubo.SetMatrix(
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, -5.1f, 0.0f)) *
		glm::scale(glm::mat4(1.0f), glm::vec3(40.0f, 0.1f, 40.0f)));
	FloorModel_ubo.UpdateOpenGLData();

	// Setup scene objects of these objects, and add them into the scene
	SceneObject floor_object;
	SceneObject corner_objects[4];
	SceneObject orbiting_object;

	floor_object.programs = &texture_program_map;
	floor_object.geometry = &geom_cube;
	floor_object.model_ubo = &FloorModel_ubo;
	floor_object.material_ubo = &WhiteMaterial_ubo;
	floor_object.texture = wood_tex;
	ObjectsInScene.push_back(floor_object);

	corner_objects[0].programs = &notexture_program_map;
	corner_objects[0].geometry = &geom_teapot;
	corner_objects[0].model_ubo = &CornerObjectsModel_ubo[0];
	corner_objects[0].material_ubo = &GreenMaterial_ubo;
	corner_objects[0].texture = 0;
	ObjectsInScene.push_back(corner_objects[0]);

	corner_objects[1].programs = &notexture_program_map;
	corner_objects[1].geometry = &geom_capsule;
	corner_objects[1].model_ubo = &CornerObjectsModel_ubo[1];
	corner_objects[1].material_ubo = &BlueMaterial_ubo;
	corner_objects[1].texture = 0;
	ObjectsInScene.push_back(corner_objects[1]);

	corner_objects[2].programs = &texture_program_map;
	corner_objects[2].geometry = &geom_torus;
	corner_objects[2].model_ubo = &CornerObjectsModel_ubo[2];
	corner_objects[2].material_ubo = &WhiteMaterial_ubo;
	corner_objects[2].texture = wood_tex;
	ObjectsInScene.push_back(corner_objects[2]);

	corner_objects[3].programs = &texture_program_map;
	corner_objects[3].geometry = &geom_cube;
	corner_objects[3].model_ubo = &CornerObjectsModel_ubo[3];
	corner_objects[3].material_ubo = &WhiteMaterial_ubo;
	corner_objects[3].texture = lenna_tex;
	ObjectsInScene.push_back(corner_objects[3]);

	orbiting_object.programs = &texture_program_map;
	orbiting_object.geometry = &geom_cube;
	orbiting_object.model_ubo = &OrbitingObjectModel_ubo;
	orbiting_object.material_ubo = &WhiteMaterial_ubo;
	orbiting_object.texture = wood_tex;
	ObjectsInScene.push_back(orbiting_object);

	// Central object's model matrix
	CentralObjectModel_ubo.Init();
	CentralObjectModel_ubo.SetMatrix(glm::mat4(1.0f));
	CentralObjectModel_ubo.UpdateOpenGLData();

	//----------------------------------------------
	//--  Prepare framebuffers and textures

	// Prepare cubemap texture into which we render the environment.
	glGenTextures(1, &Cubemap_ColorTexture);
	glBindTexture(GL_TEXTURE_CUBE_MAP, Cubemap_ColorTexture);
	SetTextureCubeParameters(GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_LINEAR, GL_LINEAR);
	for (int face = 0; face < 6; face++)
		glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, 0, GL_RGBA8, cubemap_size, cubemap_size, 0, GL_RGBA, GL_UNSIGNED_BYTE, nullptr);
	glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
	
	// Prepare also the depth cubemap texture, we need the depth test
	glGenTextures(1, &Cubemap_DepthTexture);
	glBindTexture(GL_TEXTURE_CUBE_MAP, Cubemap_DepthTexture);
	SetTextureCubeParameters(GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_LINEAR, GL_LINEAR);
	for (int face = 0; face < 6; face++)
		glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, 0, GL_DEPTH_COMPONENT, cubemap_size, cubemap_size, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_INT, nullptr);
	glBindTexture(GL_TEXTURE_CUBE_MAP, 0);

	// Create framebuffers - create six framebuffers, each with only one cubemap face
	glGenFramebuffers(6, CubemapSeparateFBO);
	for (int face = 0; face < 6; face++)
	{
		glBindFramebuffer(GL_FRAMEBUFFER, CubemapSeparateFBO[face]);
		// Use glFramebufferTexture2D to attach a single face
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, Cubemap_ColorTexture, 0);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, Cubemap_DepthTexture, 0);
		glDrawBuffers(1, DrawBuffersConstants);
		CheckFramebufferStatus("CubemapSeparateFBO");
	}
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	// Create framebuffers - create a single framebuffer with the whole cubemap
	glGenFramebuffers(1, &CubemapFBO);
	glBindFramebuffer(GL_FRAMEBUFFER, CubemapFBO);
	// Use glFramebufferTexture to attach the whole cubemap texture
	glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, Cubemap_ColorTexture, 0);
	glFramebufferTexture(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, Cubemap_DepthTexture, 0);
	glDrawBuffers(1, DrawBuffersConstants);
	CheckFramebufferStatus("CubemapFBO");
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	//----------------------------------------------
	//--  Miscelaneous

	// Create the query object
	glGenQueries(1, &RenderTimeQuery);
}

/// Updates the scene: performs animations, updates the data of the buffers, etc.
void update_scene(int app_time_diff_ms)
{
	//----------------------------------------------
	// Choose the correct skybox
	
	if (skybox_texture == SKYBOX_MEADOW)
	{
		current_skybox_tex = meadow_skybox_tex;

		// Set the light's position to correspond to the position of the sun in skybox texture
		PhongLights_ubo.PhongLights[0].position = glm::vec4(1.0f, 0.25f, 1.0f, 0.0f);
		PhongLights_ubo.UpdateOpenGLData();
	}
	else if (skybox_texture == SKYBOX_MIRAMAR)
	{
		current_skybox_tex = miramar_skybox_tex;

		// Set the light's position to correspond to the position of the sun in skybox texture
		PhongLights_ubo.PhongLights[0].position = glm::vec4(1.0f, 2.0f, -1.0f, 0.0f);
		PhongLights_ubo.UpdateOpenGLData();
	}

	//----------------------------------------------
	// Data of the main camera
	
	CameraData_ubo.SetProjection(
		glm::perspective(glm::radians(45.0f), float(win_width) / float(win_height), 0.1f, 1000.0f));
	CameraData_ubo.SetCamera(the_camera);
	CameraData_ubo.UpdateOpenGLData();

	//----------------------------------------------
	//--  Animate all objects

	// Application time in seconds
	float app_time_s = app_time_ms * 0.001f;

	glm::mat4 matrix;

	// 0. object is a teapot
	matrix =
		// Its position and orientation in the world (it rotates around its y-axis)
		glm::translate(glm::mat4(1.0f), glm::vec3(-30.0f, -5.0f, -30.0f)) *
		glm::rotate(glm::mat4(1.0f), app_time_s * 1.5f, glm::vec3(0.0f, 1.0f, 0.0f)) *
		// Its size
		glm::scale(glm::mat4(1.0f), glm::vec3(5.0f)) *
		// Its base position and orientation (so that it sits on the floor)
		glm::mat4(1.0f);
	CornerObjectsModel_ubo[0].SetMatrix(matrix);
	CornerObjectsModel_ubo[0].UpdateOpenGLData();

	// 1. object is a capsule
	matrix =
		// Its position and orientation in the world (it rotates around its y-axis)
		glm::translate(glm::mat4(1.0f), glm::vec3(30.0f, -5.0f, -30.0f)) *
		glm::rotate(glm::mat4(1.0f), app_time_s * 1.1f, glm::vec3(0.0f, 1.0f, 0.0f)) *
		// Its size
		glm::scale(glm::mat4(1.0f), glm::vec3(5.0f)) *
		// Its base position and orientation (so that it sits on the floor)
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 0.5f, 0.0f)) *
		glm::rotate(glm::mat4(1.0f), glm::radians(90.0f), glm::vec3(1.0f, 0.0f, 0.0f));
	CornerObjectsModel_ubo[1].SetMatrix(matrix);
	CornerObjectsModel_ubo[1].UpdateOpenGLData();

	// 2. object is a torus
	matrix =
		// Its position and orientation in the world (it rotates around its y-axis)
		glm::translate(glm::mat4(1.0f), glm::vec3(-30.0f, -5.0f, 30.0f)) *
		glm::rotate(glm::mat4(1.0f), app_time_s * 2.1f, glm::vec3(0.0f, 1.0f, 0.0f)) *
		// Its size
		glm::scale(glm::mat4(1.0f), glm::vec3(5.0f)) *
		// Its base position and orientation (so that it sits on the floor)
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 1.5f, 0.0f)) *
		glm::rotate(glm::mat4(1.0f), glm::radians(90.0f), glm::vec3(1.0f, 0.0f, 0.0f));
	CornerObjectsModel_ubo[2].SetMatrix(matrix);
	CornerObjectsModel_ubo[2].UpdateOpenGLData();

	// 3. object is a cube
	matrix =
		// Its position and orientation in the world (it rotates around its y-axis)
		glm::translate(glm::mat4(1.0f), glm::vec3(30.0f, -5.0f, 30.0f)) *
		glm::rotate(glm::mat4(1.0f), app_time_s * 1.9f, glm::vec3(0.0f, 1.0f, 0.0f)) *
		// Its size
		glm::scale(glm::mat4(1.0f), glm::vec3(5.0f)) *
		// Its base position and orientation (so that it sits on the floor)
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 1.0f, 0.0f));
	CornerObjectsModel_ubo[3].SetMatrix(matrix);
	CornerObjectsModel_ubo[3].UpdateOpenGLData();

	// Orbiting object
	matrix =
		// Its position and orientation in the world (it rotates around central y-axis and also chaotically around its own center)
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 10.0f, 0.0f)) *
		glm::rotate(glm::mat4(1.0f), app_time_s * 2.0f, glm::vec3(0.0f, 1.0f, 0.0f)) *
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 0.0f, 20.0f)) *
		glm::rotate(glm::mat4(1.0f), app_time_s * 1.2f, glm::vec3(1.0f, 0.0f, 0.0f)) *
		glm::rotate(glm::mat4(1.0f), app_time_s * 1.4f, glm::vec3(0.0f, 1.0f, 0.0f)) *
		glm::rotate(glm::mat4(1.0f), app_time_s * 1.6f, glm::vec3(0.0f, 0.0f, 1.0f)) *
		// Its size
		glm::scale(glm::mat4(1.0f), glm::vec3(3.0f));
	OrbitingObjectModel_ubo.SetMatrix(matrix);
	OrbitingObjectModel_ubo.UpdateOpenGLData();
}

void render_skybox_and_scene(int render_program, bool use_instancing = false)
{
	// This function renderes the scene and the skybox. It does not render the central object.
	// Parameters:
	//		render_program	.. type of the program to be used when rendering (e.g. to cube map, standard, ...)
	//		use_instancing	.. when true, each object is rendered six times using instancing

	// Render the skybox
	if (skybox_program_map.UseProgram(render_program))
	{
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_CUBE_MAP, current_skybox_tex);

		// Render the skybox only into the color buffer, do not use depth test, do not update depth buffer
		glDepthMask(GL_FALSE);
		glDisable(GL_DEPTH_TEST);

		// We render the skybox using a fullscreen quad, and we compute the view direction in vertex shader
		geom_fullscreen_quad.BindVAO();
		if (use_instancing)
			geom_fullscreen_quad.DrawInstanced(6);
		else
			geom_fullscreen_quad.Draw();
	
		// Restore the depth test and the depth buffer writes
		glDepthMask(GL_TRUE);
		glEnable(GL_DEPTH_TEST);
	}

	// Render all objects in the scene
	for (auto iter = ObjectsInScene.begin(); iter != ObjectsInScene.end(); ++iter)
	{
		if (!iter->programs->UseProgram(render_program))
			continue;		// The program is not ready, skip this object and render another one

		if (iter->material_ubo)
			iter->material_ubo->BindBuffer(DEFAULT_MATERIAL_BINDING);
		if (iter->model_ubo)
			iter->model_ubo->BindBuffer(DEFAULT_OBJECT_BINDING);

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, iter->texture);

		if (iter->geometry)
		{
			iter->geometry->BindVAO();
			if (use_instancing)
				iter->geometry->DrawInstanced(6);
			else
				iter->geometry->Draw();
		}
	}
}

/// Renders the whole frame
void render_scene()
{
	// Start measuring the elapsed time
	glBeginQuery(GL_TIME_ELAPSED, RenderTimeQuery);

	// Render into the cubemap
	if (cubemap_update_method == CUBE_MAP_UPDATE_SEPARATE_FACES)
	{
		// Render the scene six times - once for each face
		for (int face = 0; face < 6; face++)
		{
			// Set the framebuffer - set a single cubemap face
			glBindFramebuffer(GL_FRAMEBUFFER, CubemapSeparateFBO[face]);
			glViewport(0, 0, cubemap_size, cubemap_size);

			// Clear the framebuffer, clear the color and depth
			glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
			glClearDepth(1.0);
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			glEnable(GL_DEPTH_TEST);

			// Set the data of the camera and the lights
			CubemapSeparateCameraData_ubo[face].BindBuffer(DEFAULT_CAMERA_BINDING);
			PhongLights_ubo.BindBuffer(DEFAULT_LIGHTS_BINDING);

			render_skybox_and_scene(RENDER_STANDARD, false);
		}
	}
	else if (
		(cubemap_update_method == CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP) ||
		(cubemap_update_method == CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP_INVOCATIONS) ||
		(cubemap_update_method == CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP_INSTANCING) ||
		(cubemap_update_method == CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP_INSTANCING_NO_GS))
	{
		// Render the scene only once - into all faces simultaneously

		// Set the framebuffer - set all cubemap faces
		glBindFramebuffer(GL_FRAMEBUFFER, CubemapFBO);
		glViewport(0, 0, cubemap_size, cubemap_size);

		// Clear the framebuffer, clear the color and depth
		glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
		glClearDepth(1.0);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glEnable(GL_DEPTH_TEST);

		// Set the data of the camera and the lights
		CubemapCameraData_ubo.BindBuffer(DEFAULT_CAMERA_BINDING);
		PhongLights_ubo.BindBuffer(DEFAULT_LIGHTS_BINDING);

		// Render the scene
		switch (cubemap_update_method)
		{
		case CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP:
			render_skybox_and_scene(RENDER_TO_CUBE, false);
			break;
		case CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP_INVOCATIONS:
			render_skybox_and_scene(RENDER_TO_CUBE_INVOCATIONS, false);
			break;
		case CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP_INSTANCING:
			render_skybox_and_scene(RENDER_TO_CUBE_INSTANCING, true);
			break;
		case CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP_INSTANCING_NO_GS:
			render_skybox_and_scene(RENDER_TO_CUBE_INSTANCING_NO_GS, true);
			break;
		}
	}

	// Render into the main window
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glViewport(0, 0, win_width, win_height);

	// Clear the framebuffer, clear the color and depth
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);

	// Set the data of the camera and the lights
	CameraData_ubo.BindBuffer(DEFAULT_CAMERA_BINDING);
	PhongLights_ubo.BindBuffer(DEFAULT_LIGHTS_BINDING);

	// Render the scene
	render_skybox_and_scene(RENDER_STANDARD, false);

	// Render the central object
	if ((use_reflections && reflection_program.IsValid()) ||
		(!use_reflections && display_cubemap_program.IsValid()))
	{
		// Choose the proper cubemap texture: when the texture is no being updated,
		// use the loaded texture; when the texture is updated, use the update texture
		GLuint reflection_texture = (cubemap_update_method == CUBE_MAP_UPDATE_NO_UPDATE) ? current_skybox_tex : Cubemap_ColorTexture;

		// Choose the proper shader and set proper textures
		if (use_reflections)
		{
			reflection_program.Use();
			glActiveTexture(GL_TEXTURE0);		glBindTexture(GL_TEXTURE_2D, wood_tex);
			glActiveTexture(GL_TEXTURE1);		glBindTexture(GL_TEXTURE_CUBE_MAP, reflection_texture);
		}
		else
		{
			display_cubemap_program.Use();
			glActiveTexture(GL_TEXTURE0);		glBindTexture(GL_TEXTURE_CUBE_MAP, reflection_texture);
		}

		// Set object's position
		CentralObjectModel_ubo.BindBuffer(DEFAULT_OBJECT_BINDING);

		// Render proper object
		switch (central_object)
		{
		case CENTRAL_OBJECT_CUBE:		geom_cube.BindVAO();		geom_cube.Draw();		break;
		case CENTRAL_OBJECT_SPHERE:		geom_sphere.BindVAO();		geom_sphere.Draw();		break;
		case CENTRAL_OBJECT_TORUS:		geom_torus.BindVAO();		geom_torus.Draw();		break;
		case CENTRAL_OBJECT_CYLINDER:	geom_cylinder.BindVAO();	geom_cylinder.Draw();	break;
		case CENTRAL_OBJECT_CAPSULE:	geom_capsule.BindVAO();		geom_capsule.Draw();	break;
		case CENTRAL_OBJECT_TEAPOT:		geom_teapot.BindVAO();		geom_teapot.Draw();		break;
		}
	}

	// Reset the VAO and the program
	glBindVertexArray(0);
	glUseProgram(0);
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	// Stop measuring the elapsed time
	glEndQuery(GL_TIME_ELAPSED);

	// Evaluate the query
	glFinish();					// Wait for OpenGL, don't forget OpenGL is asynchronous
	GLuint64 render_time;
	glGetQueryObjectui64v(RenderTimeQuery, GL_QUERY_RESULT, &render_time);
	render_time_ms = float(render_time) * 1e-6f;
}

//-------------------
//----    GUI    ----
//-------------------

/// Callback when user clicks on Reload button
void TW_CALL reload(void *)
{
	reload_shaders();
}

void init_gui()
{
	cubemap_update_method = CUBE_MAP_UPDATE_NO_UPDATE;
	central_object = CENTRAL_OBJECT_CUBE;
	skybox_texture = SKYBOX_MEADOW;
	use_reflections = false;

	// Initialize AntTweakBar
	TwInit(TW_OPENGL_CORE, nullptr);
	TwWindowSize(win_width, win_height);
	TwGLUTModifiersFunc(glutGetModifiers);

	// Initialize GUI
	the_gui = TwNewBar("Parameters");
	TwAddButton(the_gui, "Reload", reload, nullptr, nullptr);

	// Render time (ms)
	TwAddVarRO(the_gui, "Render time (ms)", TW_TYPE_FLOAT, &render_time_ms, nullptr);

	// Skybox
	TwEnumVal skyboxes[2] = {
		{SKYBOX_MEADOW, "Meadow"},
		{SKYBOX_MIRAMAR, "Miramar"},
	};
	TwType skyboxType = TwDefineEnum("Skyboxes", skyboxes, 2);
	TwAddVarRW(the_gui, "Skybox", skyboxType, &skybox_texture, nullptr);

	// Reflections
	TwAddVarRW(the_gui, "Reflections", TW_TYPE_BOOLCPP, &use_reflections, nullptr);

	// Cube map update
	TwEnumVal updates[6] = {
		{CUBE_MAP_UPDATE_NO_UPDATE, "No update"},
		{CUBE_MAP_UPDATE_SEPARATE_FACES, "Separate faces"},
		{CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP, "Render to cube map"},
		{CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP_INVOCATIONS, "R2CM: Use invocations"},
		{CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP_INSTANCING, "R2CM: Use instancing"},
		{CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP_INSTANCING_NO_GS, "R2CM: Use instancing, no GS"},
	};
	TwType updateType = TwDefineEnum("Updates", updates, 6);
	TwAddVarRW(the_gui, "Cube map update", updateType, &cubemap_update_method, nullptr);

	// Central object
	TwEnumVal objects[6] = {
		{CENTRAL_OBJECT_CUBE, "Cube"},
		{CENTRAL_OBJECT_SPHERE, "Sphere"},
		{CENTRAL_OBJECT_TORUS, "Torus"},
		{CENTRAL_OBJECT_CYLINDER, "Cylinder"},
		{CENTRAL_OBJECT_CAPSULE, "Capsule"},
		{CENTRAL_OBJECT_TEAPOT, "Teapot"},
	};
	TwType objectType = TwDefineEnum("Objects", objects, 6);
	TwAddVarRW(the_gui, "Central object", objectType, &central_object, nullptr);
}

//---------------------------
//----    APPLICATION    ----
//---------------------------

/// GLUT callback - when the user presses a key. See glutKeyboardFunc for more info
void on_keyboard_func(unsigned char key, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventKeyboardGLUT(key, x, y))
		return;		// Already handled by AntTweakBar

	switch (key)
	{
	case 27:		// Escape key
		exit(0);
		break;
	}
}

/// GLUT callback - when the user presses a special key (like F1 key). See glutSpecialFunc for more info
void on_special_func(int key, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventSpecialGLUT(key, x, y))
		return;		// Already handled by AntTweakBar
}

/// GLUT callback - when the user presses or releases a mouse button. See glutMouseFunc for more info
void on_mouse_func(int button, int state, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseButtonGLUT(button, state, x, y))
		return;		// Already handled by AntTweakBar

	the_camera.OnMouseFunc(button, state, x, y);
}

/// GLUT callback - when the user moves the mouse when holding a botton. See glutMotionFunc for more info
void on_motion_func(int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseMotionGLUT(x, y))
		return;		// Already handled by AntTweakBar

	the_camera.OnMotionFunc(x, y);
}

/// GLUT callback - when the user moves the mouse when not holding a botton. See glutPassiveMotionFunc for more info
void on_passive_motion_func(int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseMotionGLUT(x, y))
		return;		// Already handled by AntTweakBar
}

/// GLUT callback - when the window is resized. See glutReshapeFunc for more info
void on_reshape(int width, int height)
{
	win_width = width;
	win_height = height;
	
	glViewport(0, 0, win_width, win_height);
	
	// Inform AntTweakBar
	TwWindowSize(win_width, win_height);
}

/// GLUT callback - when our animation timer requests a new frame. See glutTimerFunc for more info
void on_timer(int)
{
	// Register the timer to be called again in 20 ms, and rerender the scene
	glutTimerFunc(20, on_timer, 0);
	glutPostRedisplay();
}

/// GLUT callback - when the window needs to be redrawn. See glutDisplayFunc for more info
void on_display()
{
	//--  Update all the data

	// Update the application time
	int current_glut_time = glutGet(GLUT_ELAPSED_TIME);
	int app_time_diff_ms = current_glut_time - last_glut_time;
	app_time_ms += app_time_diff_ms;
	last_glut_time = current_glut_time;

	// Update the scene
	update_scene(app_time_diff_ms);

	//--  Render the scene

	// Render the scene
	render_scene();
	
	// Render the GUI
	TwDraw();

	// Swaps the front and back buffer (double-buffering)
	glutSwapBuffers();
}

/// Callback function to be called when we make an error in OpenGL
void GLAPIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const char* message, const void* userParam)
{
	switch (type)			// Report only some errors
	{
	case GL_DEBUG_TYPE_ERROR:
	case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
	case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
		cout << message << endl;		// Put the breakpoint here
		return;
	default:
		return;
	}
}

/// C main function :-)
int main(int argc, char ** argv)
{
	// Initialize GLUT
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);		// Return from the main loop

	// Set OpenGL Context parameters
	glutInitContextVersion(4, 3);				// Use OpenGL 4.3
	glutInitContextProfile(GLUT_CORE_PROFILE);	// Use OpenGL core profile
	glutInitContextFlags(GLUT_DEBUG);			// Use OpenGL debug context

	// Initialize window
	glutInitWindowSize(win_width, win_height);
	glutCreateWindow("PV227 - Cv8: Environment mapping, render to cube map");

	// Initialize GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	// Initialize DevIL library
	InitDevIL();

	// Set the OpenGL debug callback
	SetDebugCallback(simple_debug_callback);

	// Register GLUT callbacks
	glutDisplayFunc(on_display);
	glutReshapeFunc(on_reshape);
	glutTimerFunc(20, on_timer, 0);

	glutKeyboardFunc(on_keyboard_func);
	glutSpecialFunc(on_special_func);
	glutMouseFunc(on_mouse_func);
	glutMotionFunc(on_motion_func);
	glutPassiveMotionFunc(on_passive_motion_func);

	// Initialize OpenGL stuff
	init_gui();
	init_scene();

	// Everything is loaded and prepared, reset the application time
	last_glut_time = glutGet(GLUT_ELAPSED_TIME);
	app_time_ms = 0;

	// Run the main loop
	glutMainLoop();

	// Unload AntTweakBar
	TwTerminate();

	return 0;
}