// PV227, Lesson 8: Environment mapping, render to cube map
#pragma once

#include "../../Framework/PV227.h"

using namespace std;
using namespace PV227;

//---------------------
//----    SCENE    ----
//---------------------

// Shader programs
ShaderProgram notexture_program;
ShaderProgram texture_program;
ShaderProgram skybox_program;
ShaderProgram skybox_to_cube_program;
ShaderProgram texture_to_cube_program;
ShaderProgram notexture_to_cube_program;
ShaderProgram skybox_to_cube_invocations_program;
ShaderProgram texture_to_cube_invocations_program;
ShaderProgram notexture_to_cube_invocations_program;
ShaderProgram skybox_to_cube_instancing_program;
ShaderProgram texture_to_cube_instancing_program;
ShaderProgram notexture_to_cube_instancing_program;
ShaderProgram skybox_to_cube_instancing_no_gs_program;
ShaderProgram texture_to_cube_instancing_no_gs_program;
ShaderProgram notexture_to_cube_instancing_no_gs_program;
ShaderProgram display_cubemap_program;
ShaderProgram reflection_program;

// Shader program maps and map keys
ShaderProgramMap notexture_program_map;
ShaderProgramMap texture_program_map;
ShaderProgramMap skybox_program_map;
const int RENDER_STANDARD = 0;
const int RENDER_TO_CUBE = 1;
const int RENDER_TO_CUBE_INVOCATIONS = 2;
const int RENDER_TO_CUBE_INSTANCING = 3;
const int RENDER_TO_CUBE_INSTANCING_NO_GS = 4;

// Geometries we use in this lecture
Geometry geom_cube;
Geometry geom_sphere;
Geometry geom_torus;
Geometry geom_cylinder;
Geometry geom_capsule;
Geometry geom_teapot;
Geometry geom_fullscreen_quad;

// List of geometries which we choose for our scene
std::vector<std::pair<Geometry *, glm::mat4> > Geometries;

glm::mat4 CornerObjectsLocalMatrices[4];
glm::mat4 OrbitingObjectLocalMatrix;

// A camera that allows us to look at the object from different views
SimpleCamera the_camera;

// Textures
GLuint wood_tex;
GLuint lenna_tex;
GLuint meadow_skybox_tex;
GLuint miramar_skybox_tex;

// List of textures which we choose for our scene
std::vector<GLuint> Textures;
GLuint current_skybox_tex;

// Data of our materials
MaterialData_UBO RedMaterial_ubo;
MaterialData_UBO GreenMaterial_ubo;
MaterialData_UBO BlueMaterial_ubo;
MaterialData_UBO CyanMaterial_ubo;
MaterialData_UBO MagentaMaterial_ubo;
MaterialData_UBO YellowMaterial_ubo;
MaterialData_UBO WhiteMaterial_ubo;
MaterialData_UBO FloorMaterial_ubo;
// List of materials which we choose for our scene
std::vector<MaterialData_UBO *> Colors_ubo;

// Data of the our objects (their position)
std::vector<ModelData_UBO> Models_ubo;
ModelData_UBO FloorModel_ubo;
ModelData_UBO CornerObjectsModel_ubo[4];
ModelData_UBO OrbitingObjectModel_ubo;
ModelData_UBO CentralObjectModel_ubo;

// List of objects in the scene. Each object is defined by shaders it uses, its material,
// model matrix, texture (if used), and geometry.
struct SceneObject
{
	ShaderProgramMap *programs;			// Map of shader programs to use
	Geometry *geometry;					// Geomety of the object
	ModelData_UBO *model_ubo;			// Model matrix with the position of the object
	MaterialData_UBO *material_ubo;		// Material of the object
	GLuint texture;						// Texture of the object (or 0 if no texture is used)
};
std::vector<SceneObject> ObjectsInScene;

// Data of all our lights - positions, colors, etc.
PhongLightsData_UBO PhongLights_ubo;

// Data of our camera - view matrix, projection matrix, etc.
CameraData_UBO CameraData_ubo;
// Data of six cameras that are used when updating the cubemap
CameraData_UBO CubemapCameraData_ubo;				// All in one buffer
CameraData_UBO CubemapSeparateCameraData_ubo[6];	// Each in separate buffer

// OpenGL query object to get render time of one frame
GLuint RenderTimeQuery;

const int cubemap_size = 1024;
GLuint CubemapFBO, CubemapSeparateFBO[6];
GLuint Cubemap_ColorTexture;
GLuint Cubemap_DepthTexture;

// Functions that works with scene objects
void reload_shaders();
void init_scene();
void update_scene(int app_time_diff_ms);
void render_scene();

//-------------------
//----    GUI    ----
//-------------------

// Main AntTweakBar object
TwBar *the_gui;

// Variables that are changed with GUI
float render_time_ms;		// Read-only
int skybox_texture;
const int SKYBOX_MEADOW = 0;
const int SKYBOX_MIRAMAR = 1;

int cubemap_update_method;
const int CUBE_MAP_UPDATE_NO_UPDATE = 0;
const int CUBE_MAP_UPDATE_SEPARATE_FACES = 1;
const int CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP = 2;
const int CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP_INVOCATIONS = 3;
const int CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP_INSTANCING = 4;
const int CUBE_MAP_UPDATE_RENDER_TO_CUBE_MAP_INSTANCING_NO_GS = 5;

int central_object;
const int CENTRAL_OBJECT_CUBE = 0;
const int CENTRAL_OBJECT_SPHERE = 1;
const int CENTRAL_OBJECT_TORUS = 2;
const int CENTRAL_OBJECT_CYLINDER = 3;
const int CENTRAL_OBJECT_CAPSULE = 4;
const int CENTRAL_OBJECT_TEAPOT = 5;

bool use_reflections;

// Callbacks from the GUI
void TW_CALL reload(void *);

// Functions that works with GUI
void init_gui();

//---------------------------
//----    APPLICATION    ----

// Current window size
int win_width = 640;
int win_height = 480;

// Current time of the application in milliseconds, for animations
int app_time_ms = 0;
int last_glut_time = 0;
