#version 430 core

const vec2 tex_coords[3] = vec2[3] (
	vec2(0.0, 0.0),
	vec2(2.0, 0.0),
	vec2(0.0, 2.0)
);

out VertexData
{
	vec3 tex_coord;
} outData;

// Data of the camera
layout (std140, binding = 0) uniform CameraData
{
	mat4 projection;		// Projection matrix
	mat4 projection_inv;	// Inverse of the projection matrix
	mat4 view;				// View matrix
	mat4 view_inv;			// Inverse of the view matrix
	mat3 view_it;			// Inverse of the transpose of the top-left part 3x3 of the view matrix
	vec3 eye_position;		// Position of the eye in world space
};

//-----------------------------------------------------------------------

void main()
{
	// gl_VertexID holds the index of the vertex that is being processed
	vec2 tex_coord = tex_coords[gl_VertexID];

	// We compute the view direction based on the position of the vertex in clip space
	vec4 pos_cs = vec4(tex_coord * 2.0 - 1.0, 0.0, 1.0);	// Position of the vertex in clip space
	vec4 pos_vs = projection_inv * pos_cs;					// Use inverse of the projection matrix to transform the position into view space
	vec3 dir_vs = pos_vs.xyz;								// In view space, the position of the vertex corresponds to the direction of the view.
	vec3 dir_ws = mat3(view_inv) * dir_vs;					// Transform the direction into the world space
	outData.tex_coord = dir_ws;								// The direction in world space is the texture coordinate for the cube map

	gl_Position = pos_cs;

	// Task 2 (optional): Do nothing, just see how the tex_coord and gl_Position is computed (rewrite it into other shaders)
}
