#version 430 core

// Input variables
layout (location = 0) in vec4 position;
layout (location = 1) in vec3 normal;
layout (location = 2) in vec2 tex_coord;

// Output variables
out VertexData
{
	vec3 normal_ws;		// Normal in world space
	vec3 position_ws;	// Position in world space
	vec2 tex_coord;
} outData;

// Data of the object
layout (std140, binding = 2) uniform ModelData
{
	mat4 model;			// Model matrix
	mat4 model_inv;		// Inverse of the model matrix
	mat3 model_it;		// Inverse of the transpose of the top-left part 3x3 of the model matrix
};

//-----------------------------------------------------------------------

void main()
{
	// Compute the data of the vertex
	outData.tex_coord = tex_coord;
	outData.position_ws = vec3(model * position);
	outData.normal_ws = normalize(model_it * normal);

	// Task 2: Actually, nothing to do here, just notice that the vertex shader computes only the tex_coord
	//		and the data in world space, because they do not depend on the camera. It also does not compute
	//		gl_Position, it is computed in geometry shader.
}
